@@ -5,7 +5,7 @@ import operator
 import sys
 import unittest
 from collections import namedtuple
-from datetime import datetime, timedelta
+from datetime import datetime, timedelta, MINYEAR
 
 import numpy as np
 import pytest
@@ -17,7 +17,7 @@ from cftime import real_datetime
 from cftime import (DateFromJulianDay, Datetime360Day, DatetimeAllLeap,
                     DatetimeGregorian, DatetimeJulian, DatetimeNoLeap,
                     DatetimeProlepticGregorian, JulianDayFromDate, _parse_date,
-                    date2index, date2num, num2date, utime)
+                    date2index, date2num, num2date, utime, UNIT_CONVERSION_FACTORS)
 
 try:
     from datetime import timezone
@@ -51,6 +51,19 @@ est = timezone(timedelta(hours=-5), 'UTC')
 dtime = namedtuple('dtime', ('values', 'units', 'calendar'))
 dateformat =  '%Y-%m-%d %H:%M:%S'
 
+calendars=['standard', 'gregorian', 'proleptic_gregorian', 'noleap', 'julian',\
+           'all_leap', '365_day', '366_day', '360_day']
+def adjust_calendar(calendar):
+    # check for and remove calendar synonyms.
+    calendar = calendar.lower()
+    if calendar == 'gregorian' or calendar == 'standard':
+        return 'gregorian'
+    elif calendar == 'noleap' or calendar == '365_day':
+        return 'noleap'
+    elif calendar == 'all_leap' or calendar == '366_day':
+        return 'all_leap'
+    else:
+        return calendar
 
 class CFTimeVariable(object):
     '''dummy "netCDF" variable to hold time info'''
@@ -115,11 +128,8 @@ class cftimeTestCase(unittest.TestCase):
         t1 = self.cdftime_mixed.date2num(d1)
         d2 = datetime(1582, 10, 4, 18, tzinfo=est)
         t2 = self.cdftime_mixed.date2num(d2)
-        d3 = d2.replace(tzinfo=None)
-        t3 = self.cdftime_mixed.date2num(d3)
         assert_almost_equal(t1, 13865687.0)
-        assert_almost_equal(t2, 13865687.0)
-        assert_almost_equal(t3, 13865682.0)
+        assert_almost_equal(t2, 13865682.0) # est is 5 hours behind utc
 
     def test_tz_naive(self):
         """testing cftime"""
@@ -251,7 +261,7 @@ class cftimeTestCase(unittest.TestCase):
         # check date2num,num2date methods.
         # use datetime from cftime, since this date doesn't
         # exist in "normal" calendars.
-        d = datetimex(2000, 2, 30)
+        d = datetimex(2000, 2, 30, calendar='')
         t1 = self.cdftime_360day.date2num(d)
         assert_almost_equal(t1, 360 * 400.)
         d2 = self.cdftime_360day.num2date(t1)
@@ -280,11 +290,17 @@ class cftimeTestCase(unittest.TestCase):
         date = self.cdftime_jul.num2date(t)
         self.assertTrue(str(d) == str(date))
         # test julian day from date, date from julian day
-        d = datetime(1858, 11, 17)
-        mjd = JulianDayFromDate(d)
-        assert_almost_equal(mjd, 2400000.5)
-        date = DateFromJulianDay(mjd)
-        self.assertTrue(str(date) == str(d))
+        d = cftime.datetime(1858, 11, 17, calendar='standard')
+        # toordinal should produce same result as JulianDayFromDate
+        mjd1 = d.toordinal(fractional=True)
+        mjd2 = JulianDayFromDate(d)
+        assert_almost_equal(mjd1, 2400000.5)
+        assert_almost_equal(mjd1,mjd2)
+        # fromordinal should produce the same result as DateFromJulianDay
+        date1 = DateFromJulianDay(mjd1)
+        date2 = cftime.datetime.fromordinal(mjd1)
+        self.assertTrue(str(date1) == str(d))
+        self.assertTrue(str(date1) == str(date2))
         # test iso 8601 units string
         d = datetime(1970, 1, 1, 1)
         t = self.cdftime_iso.date2num(d)
@@ -300,7 +316,7 @@ class cftimeTestCase(unittest.TestCase):
             t2 = date2num(d, units='days since 0001-01-01 00:00:00')
             assert(abs(t2 - t) < 1e-5)  # values should be less than second
         # Check equality testing
-        d1 = datetimex(1979, 6, 21, 9, 23, 12)
+        d1 = datetimex(1979, 6, 21, 9, 23, 12, calendar='standard')
         d2 = datetime(1979, 6, 21, 9, 23, 12)
         assert(d1 == d2)
         # check timezone offset
@@ -336,7 +352,7 @@ class cftimeTestCase(unittest.TestCase):
         # check datetime hash
         d1 = datetimex(1995, 1, 1)
         d2 = datetime(1995, 1, 1)
-        d3 = datetimex(2001, 2, 30)
+        d3 = datetimex(2001, 2, 20)
         assert hash(d1) == hash(d1)
         assert hash(d1) == hash(d2)
         assert hash(d1) != hash(d3)
@@ -374,108 +390,55 @@ class cftimeTestCase(unittest.TestCase):
 
         # test rountrip accuracy
         # also tests error found in issue #349
-        calendars=['standard', 'gregorian', 'proleptic_gregorian', 'noleap', 'julian',\
-                   'all_leap', '365_day', '366_day', '360_day']
         dateref = datetime(2015,2,28,12)
-        ntimes = 1001
         verbose = True # print out max error diagnostics
+        ntimes = 101
+        def roundtrip(delta,eps,units):
+            times1 = date2num(dateref,units,calendar=calendar)
+            times1 += delta*np.arange(0,ntimes)
+            dates1 = num2date(times1,units,calendar=calendar)
+            times2 = date2num(dates1,units,calendar=calendar)
+            dates2 = num2date(times2,units,calendar=calendar)
+            err = np.abs(times1 - times2)
+            assert(np.all(err<eps))
+            dates1 = [date.strftime(dateformat) for date in dates1]
+            dates2 = [date.strftime(dateformat) for date in dates2]
+            assert(dates1==dates2)
+            return err.max()
         for calendar in calendars:
-            eps = 100.
+            eps = 1.; delta = 1.
             units = 'microseconds since 2000-01-30 01:01:01'
-            microsecs1 = date2num(dateref,units,calendar=calendar)
-            maxerr = 0
-            for n in range(ntimes):
-                microsecs1 += 1.
-                date1 = num2date(microsecs1, units, calendar=calendar)
-                microsecs2 = date2num(date1, units, calendar=calendar)
-                date2 = num2date(microsecs2, units, calendar=calendar)
-                err = np.abs(microsecs1 - microsecs2)
-                maxerr = max(err,maxerr)
-                assert(err < eps)
-                assert(date1.strftime(dateformat) == date2.strftime(dateformat))
+            maxerr = roundtrip(eps,delta,units)
             if verbose:
                 print('calendar = %s max abs err (microsecs) = %s eps = %s' % \
                      (calendar,maxerr,eps))
-            units = 'milliseconds since 1800-01-30 01:01:01'
-            eps = 0.1
-            millisecs1 = date2num(dateref,units,calendar=calendar)
-            maxerr = 0.
-            for n in range(ntimes):
-                millisecs1 += 0.001
-                date1 = num2date(millisecs1, units, calendar=calendar)
-                millisecs2 = date2num(date1, units, calendar=calendar)
-                date2 = num2date(millisecs2, units, calendar=calendar)
-                err = np.abs(millisecs1 - millisecs2)
-                maxerr = max(err,maxerr)
-                assert(err < eps)
-                assert(date1.strftime(dateformat) == date2.strftime(dateformat))
+            eps = 0.001; delta = 0.001
+            units = 'milliseconds since 2000-01-30 01:01:01'
+            maxerr = roundtrip(eps,delta,units)
             if verbose:
                 print('calendar = %s max abs err (millisecs) = %s eps = %s' % \
                      (calendar,maxerr,eps))
-            eps = 1.e-3
+            eps = 1.e-5; delta = 0.1
             units = 'seconds since 0001-01-30 01:01:01'
-            secs1 = date2num(dateref,units,calendar=calendar)
-            maxerr = 0.
-            for n in range(ntimes):
-                secs1 += 0.1
-                date1 = num2date(secs1, units, calendar=calendar)
-                secs2 = date2num(date1, units, calendar=calendar)
-                date2 = num2date(secs2, units, calendar=calendar)
-                err = np.abs(secs1 - secs2)
-                maxerr = max(err,maxerr)
-                assert(err < eps)
-                assert(date1.strftime(dateformat) == date2.strftime(dateformat))
+            maxerr = roundtrip(eps,delta,units)
             if verbose:
                 print('calendar = %s max abs err (secs) = %s eps = %s' % \
                      (calendar,maxerr,eps))
-            eps = 1.e-5
+            eps = 1.e-6; delta = 0.01
             units = 'minutes since 0001-01-30 01:01:01'
-            mins1 = date2num(dateref,units,calendar=calendar)
-            maxerr = 0.
-            for n in range(ntimes):
-                mins1 += 0.01
-                date1 = num2date(mins1, units, calendar=calendar)
-                mins2 = date2num(date1, units, calendar=calendar)
-                date2 = num2date(mins2, units, calendar=calendar)
-                err = np.abs(mins1 - mins2)
-                maxerr = max(err,maxerr)
-                assert(err < eps)
-                diff = abs(date1-date2)
-                assert(diff.microseconds < 100)
+            maxerr = roundtrip(eps,delta,units)
             if verbose:
                 print('calendar = %s max abs err (mins) = %s eps = %s' % \
                      (calendar,maxerr,eps))
-            eps = 1.e-6
+            eps = 1.e-8; delta = 0.001
             units = 'hours since 0001-01-30 01:01:01'
-            hrs1 = date2num(dateref,units,calendar=calendar)
-            maxerr = 0.
-            for n in range(ntimes):
-                hrs1 += 0.001
-                date1 = num2date(hrs1, units, calendar=calendar)
-                hrs2 = date2num(date1, units, calendar=calendar)
-                date2 = num2date(hrs2, units, calendar=calendar)
-                err = np.abs(hrs1 - hrs2)
-                maxerr = max(err,maxerr)
-                assert(err < eps)
-                diff = abs(date1-date2)
-                assert(diff.microseconds < 100)
+            maxerr = roundtrip(eps,delta,units)
             if verbose:
                 print('calendar = %s max abs err (hours) = %s eps = %s' % \
                      (calendar,maxerr,eps))
-            eps = 1.e-8
+            eps = 1.e-9; delta = 0.00001
             units = 'days since 0001-01-30 01:01:01'
-            days1 = date2num(dateref,units,calendar=calendar)
-            maxerr = 0.
-            for n in range(ntimes):
-                days1 += 0.00001
-                date1 = num2date(days1, units, calendar=calendar)
-                days2 = date2num(date1, units, calendar=calendar)
-                date2 = num2date(days2, units, calendar=calendar)
-                err = np.abs(days1 - days2)
-                maxerr = max(err,maxerr)
-                assert(err < eps)
-                diff = abs(date1-date2)
-                assert(diff.microseconds < 100)
+            maxerr = roundtrip(eps,delta,units)
             if verbose:
                 print('calendar = %s max abs err (days) = %s eps = %s' % \
                      (calendar,maxerr,eps))
@@ -697,6 +660,20 @@ class cftimeTestCase(unittest.TestCase):
         d2 = real_datetime(1582,10,4,12)
         assert (d1.dayofwk == d2.dayofwk == 0)
         assert (d1.dayofyr == d2.dayofyr == 277)
+
+        # issue 173: Return the default values of dayofwk and dayofyr
+        # when calendar is ''
+        d1 = datetimex(1582, 10, 4, 12, calendar='')
+        assert (d1.dayofwk == d1.dayofyr == -1)
+        d1 = datetimex(2020, 5, 20, calendar='')
+        assert (d1.dayofwk == d1.dayofyr == -1)
+        d1 = datetimex(2020, 5, 20, dayofwk=-2, dayofyr=-3, calendar='')
+        assert (d1.dayofwk == -2)
+        assert (d1.dayofyr == -3)
+        d1 = datetimex(2020, 5, 20, dayofwk=8, dayofyr=9, calendar='')
+        assert (d1.dayofwk == 8)
+        assert (d1.dayofyr == 9)
+        
         # issue 71: negative reference years
         # https://coastwatch.pfeg.noaa.gov/erddap/convert/time.html
         # gives 2446433 (365 days more - is it counting year 0?)
@@ -729,6 +706,121 @@ class cftimeTestCase(unittest.TestCase):
         d = datetime.strptime('2018-01-23 09:27:10.950000',"%Y-%m-%d %H:%M:%S.%f")
         units = 'seconds since 2018-01-23 09:31:42.94'
         assert(cftime.date2num(d, units) == -271.99)
+        # issue 143 - same answer for arrays vs scalars.
+        units = 'seconds since 1970-01-01 00:00:00'
+        times_in = [1261440000.0, 1261440001.0, 1261440002.0, 1261440003.0,
+                    1261440004.0, 1261440005.0]
+        times_out1 = cftime.num2date(times_in, units)
+        times_out2 = []
+        for time_in in times_in:
+            times_out2.append(cftime.num2date(time_in, units))
+        dates1 = [str(d) for d in times_out1]
+        dates2 = [str(d) for d in times_out2]
+        assert(dates1 == dates2)
+        # issue #143 formatting of microseconds
+        d = cftime.num2date(1261440000.015625,units)
+        # on windows only 100 ms precision
+        assert(str(d)[0:24] == '2009-12-22 00:00:00.0156')
+        # issue #152 add isoformat()
+        assert(d.isoformat()[0:24] == '2009-12-22T00:00:00.0156')
+        assert(d.isoformat(sep=' ')[0:24] == '2009-12-22 00:00:00.0156')
+        assert(d.isoformat(sep=' ',timespec='milliseconds') == '2009-12-22 00:00:00.015')
+        assert(d.isoformat(sep=' ',timespec='seconds') == '2009-12-22 00:00:00')
+        assert(d.isoformat(sep=' ',timespec='minutes') == '2009-12-22 00:00')
+        assert(d.isoformat(sep=' ',timespec='hours') == '2009-12-22 00')
+        # issue #165: make sure python datetime returned
+        d=num2date(0,units="seconds since 2000-01-01 00:00:00",only_use_cftime_datetimes=False)
+        assert isinstance(d, datetime)
+        # issue #169: cftime.datetime has no calendar attribute, causing dayofwk,dayofyr methods
+        # to fail.
+        c = cftime.datetime(*cftime._parse_date('7480-01-01 00:00:00'),calendar='standard')
+        assert(c.strftime() == '7480-01-01 00:00:00')
+        # issue #175: masked values not treated properly in num2date
+        times = np.ma.masked_array([-3956.7499999995343,-999999999999],mask=[False,True])
+        units='days since 1858-11-17 00:00:00'
+        dates = num2date(times, units=units, calendar='standard',\
+        only_use_cftime_datetimes=False, only_use_python_datetimes=True)
+        test = dates == np.ma.masked_array([datetime(1848, 1, 17, 6, 0, 0, 40), None],mask=[0,1])
+        assert(test.all())
+        dates = num2date(times, units=units, calendar='standard')
+        assert(str(dates)=="[cftime.DatetimeGregorian(1848, 1, 17, 6, 0, 0, 40) --]")
+#  check that time range of 200,000 + years can be represented accurately
+        calendar='standard'
+        _MAX_INT64 = np.iinfo("int64").max
+        refdate = DatetimeGregorian(292277,10,24,0,0,1)
+        for unit in ['microseconds','milliseconds','seconds']:
+            units = '%s since 01-01-01' % unit
+            time = 292471*365*86400*(1000000//int(UNIT_CONVERSION_FACTORS[unit])) + 1000000//int(UNIT_CONVERSION_FACTORS[unit])
+            date = num2date(time,units,calendar=calendar)
+            assert(date == refdate)
+            # check round-trip
+            time2 = date2num(date,units,calendar=calendar)
+            date2 = num2date(time2,units,calendar=calendar)
+            # windows doesn't have a longdouble type (it defaults to float64)
+            if np.finfo(np.longdouble).precision == 18:
+                assert(date2 == refdate)
+# microsecond roundtrip accuracy preserved over time ranges of 286 years
+# (float64 can only represent integers exactly up to 2**53-1)
+        refdate=DatetimeGregorian(286,6,3,23,47,34,740992)
+        for unit in ['microseconds','milliseconds','seconds','hours','days']:
+            units = '%s since 01-01-01' % unit
+            time = (2**53 - 1)*(1/UNIT_CONVERSION_FACTORS[unit]) + 1/UNIT_CONVERSION_FACTORS[unit]
+            date = num2date(time,units,calendar=calendar)
+            assert(date == refdate)
+            # check round-trip
+            time2 = date2num(date,units,calendar=calendar)
+            date2 = num2date(time2,units,calendar=calendar)
+            assert(date2 == refdate)
+# issue #185: date2num should work the numpy scalar array of dates (1.2.0 regression)
+        dates = np.array(datetime(2010, 2, 2, 0, 0))
+        assert (date2num(dates, units="hours since 2010-02-01 00:00:00") == 24.)
+# issue #187 - roundtrip near second boundary
+        dt1 = datetime(1810, 4, 24, 16, 15, 10)
+        units = 'days since -4713-01-01 12:00'
+        dt2 = num2date(date2num(dt1, units), units, calendar='proleptic_gregorian')
+        dt2 = num2date(date2num(dt1, units, calendar='standard'), units)
+        assert(dt1 == dt2)
+# issue #189 - leap years calculated incorrectly for negative years in proleptic_gregorian calendar
+        dt1 = datetime(2020, 4, 24, 16, 15, 10)
+        units = 'days since -4713-01-01 12:00'
+        cal = 'proleptic_gregorian'
+        dt2 = num2date(date2num(dt1, units, cal), units, cal)
+        assert(dt1 == dt2)
+# issue #198 - cftime.datetime creates calendar specific datetimes that
+# support addition/subtraction of timedeltas.
+        for cal in calendars:
+            dt = cftime.datetime(2020, 1, 1, calendar=cal)
+            dt += timedelta(hours=1)
+            assert(str(dt) == '2020-01-01 01:00:00')
+# issue #193 - years with more than four digits in reference date
+        assert(cftime.date2num(cftime.datetime(18000, 12, 1, 0, 0), 'days since 18000-1-1', '360_day') == 330.0)
+        # julian day not including year zero
+        d = cftime.datetime(2020, 12, 1, 12, calendar='julian')
+        units = 'days since -4713-1-1 12:00'
+        jd = cftime.date2num(d,units,calendar='julian')
+        assert(jd == 2459198.0)
+        # if calendar=None, use input date to determine calendar
+        jd = cftime.date2num(d,units,calendar=None)
+        assert(jd == 2459198.0)
+        # if no calendar specified, use calendar associated with datetime
+        # instance.
+        jd = cftime.date2num(d,units)
+        assert(jd == 2459198.0)
+        # use 'standard' calendar
+        jd = cftime.date2num(d,units,calendar='standard')
+        assert(jd == 2459185.0)
+
+        # issue #211
+        # (masked array handling in date2num - AttributeError:
+        # 'cftime._cftime.DatetimeGregorian' object has no attribute 'view')
+        m = np.ma.asarray(
+            [cftime.DatetimeGregorian(2014, 8, 1, 12, 0, 0, 0)]
+            )
+        assert(
+             cftime.date2num(m, units="seconds since 2000-1-1")==[4.602096e+08]
+              )
+
+
 
 class TestDate2index(unittest.TestCase):
 
@@ -1110,6 +1202,10 @@ class DateTime(unittest.TestCase):
         self.assertEqual(self.date1_365_day.replace(minute=3).minute, 3)
         self.assertEqual(self.date1_365_day.replace(second=3).second, 3)
         self.assertEqual(self.date1_365_day.replace(microsecond=3).microsecond, 3)
+        d = datetimex(2000, 2, 3, calendar='noleap')
+        self.assertEqual(d.replace(year=1999).calendar, 'noleap')
+        d = datetimex(2000, 2, 3, calendar='')
+        self.assertEqual(d.replace(year=1999).calendar, '')
 
     def test_pickling(self):
         "Test reversibility of pickling."
@@ -1223,19 +1319,19 @@ class DateTime(unittest.TestCase):
         for op, expected in [(operator.gt, '__lt__'), (operator.ge, '__le__'),
                              (operator.eq, '__eq__'), (operator.ne, '__ne__'),
                              (operator.lt, '__gt__'), (operator.le, '__ge__')]:
-            with self.assertRaisesRegexp(NotImplementedError, expected):
+            with self.assertRaisesRegex(NotImplementedError, expected):
                 op(self.date1_365_day, Rich())
 
-            with self.assertRaisesRegexp(NotImplementedError, '__richcmp__'):
+            with self.assertRaisesRegex(NotImplementedError, '__richcmp__'):
                 op(self.date1_365_day, CythonRich())
 
         # Test RHS operand comparison operator processing.
         for op in [operator.gt, operator.ge, operator.eq, operator.ne,
                    operator.lt, operator.le]:
-            with self.assertRaisesRegexp(TypeError, 'cannot compare'):
+            with self.assertRaisesRegex(TypeError, 'cannot compare'):
                 op(Pass(), self.date1_365_day)
 
-            with self.assertRaisesRegexp(TypeError, 'cannot compare'):
+            with self.assertRaisesRegex(TypeError, 'cannot compare'):
                 op(Pass___cmp__(), self.date1_365_day)
 
 
@@ -1286,13 +1382,13 @@ class issue57TestCase(unittest.TestCase):
         for datestr in ("days since2017-05-01 ", "dayssince 2017-05-01 00:00", "days snce 2017-05-01 00:00", "days_since_2017-05-01 00:00",
             "days_since_2017-05-01_00:00"):
             self.assertRaises(
-                ValueError, cftime._cftime._dateparse, datestr)
+                ValueError, cftime._cftime._dateparse, datestr, 'standard')
 
             self.assertRaises(
-                ValueError, cftime._cftime.num2date, 1, datestr)
+                ValueError, cftime._cftime.num2date, 1, datestr, 'standard')
 
             self.assertRaises(
-                ValueError, cftime._cftime.date2num, datetime(1900, 1, 1, 0), datestr)
+                ValueError, cftime._cftime.date2num, datetime(1900, 1, 1, 0), datestr, 'standard')
 
 
 _DATE_TYPES = [DatetimeNoLeap, DatetimeAllLeap, DatetimeJulian, Datetime360Day,
@@ -1459,7 +1555,8 @@ def test_num2date_only_use_cftime_datetimes_negative_years(
         calendar, expected_date_type):
     result = num2date(-1000., units='days since 0001-01-01', calendar=calendar,
                       only_use_cftime_datetimes=True)
-    assert isinstance(result, expected_date_type)
+    assert isinstance(result, datetimex)
+    assert (result.calendar == adjust_calendar(calendar))
 
 
 @pytest.mark.parametrize(
@@ -1470,7 +1567,8 @@ def test_num2date_only_use_cftime_datetimes_pre_gregorian(
         calendar, expected_date_type):
     result = num2date(1., units='days since 0001-01-01', calendar=calendar,
                       only_use_cftime_datetimes=True)
-    assert isinstance(result, expected_date_type)
+    assert isinstance(result, datetimex)
+    assert (result.calendar == adjust_calendar(calendar))
 
 
 @pytest.mark.parametrize(
@@ -1481,12 +1579,15 @@ def test_num2date_only_use_cftime_datetimes_post_gregorian(
         calendar, expected_date_type):
     result = num2date(0., units='days since 1582-10-15', calendar=calendar,
                       only_use_cftime_datetimes=True)
-    assert isinstance(result, expected_date_type)
+    assert isinstance(result, datetimex)
+    assert (result.calendar == adjust_calendar(calendar))
 
 
 def test_repr():
-    expected = 'cftime.datetime(2000-01-01 00:00:00)'
-    assert repr(datetimex(2000, 1, 1)) == expected
+    expected = "cftime.datetime(2000, 1, 1, 0, 0, 0, 0, calendar='gregorian')"
+    assert repr(datetimex(2000, 1, 1, calendar='standard')) == expected
+    expected = "cftime.datetime(2000, 1, 1, 0, 0, 0, 0, calendar='')"
+    assert repr(datetimex(2000, 1, 1, calendar=None)) == expected
 
 
 def test_dayofyr_after_replace(date_type):
@@ -1519,5 +1620,320 @@ def test_replace_dayofyr_or_dayofwk_error(date_type, argument):
         date_type(1, 1, 1).replace(**{argument: 3})
 
 
+def test_dayofyr_after_timedelta_addition(date_type):
+    initial_date = date_type(1, 1, 2)
+    date_after_timedelta_addition = initial_date + timedelta(days=1)
+    assert initial_date.dayofyr == 2
+    assert date_after_timedelta_addition.dayofyr == 3
+
+
+def test_exact_datetime_difference(date_type):
+    b = date_type(2000, 1, 2, 0, 0, 0, 5)
+    a = date_type(2000, 1, 2)
+    result = b - a
+    expected = timedelta(microseconds=5)
+    assert result == expected
+
+
+_SHAPES = [(4, ), (2, 2)]
+_MICROSECOND_UNITS = ["microseconds", "microsecond", "microsec", "microsecs"]
+_MILLISECOND_UNITS = ["milliseconds", "millisecond", "millisec", "millisecs", "ms"]
+_SECOND_UNITS = ["seconds", "second", "sec", "secs", "s"]
+_MINUTE_UNITS = ["minutes", "minute", "min", "mins"]
+_HOUR_UNITS = ["hours", "hour", "hr", "hrs", "h"]
+_DAY_UNITS = ["day", "days", "d"]
+_MONTH_UNITS = ["month", "months"]
+_DTYPES = [np.dtype("int64"), np.dtype("float64")]
+_STANDARD_CALENDARS = [
+    "standard",
+    "gregorian",
+    "proleptic_gregorian"
+]
+_REAL_WORLD_CALENDARS = [
+    "julian",
+    "standard",
+    "gregorian",
+    "proleptic_gregorian"
+]
+_ARTIFICIAL_CALENDARS = ["noleap", "all_leap", "360_day"]
+
+
+@pytest.fixture(params=_SHAPES)
+def shape(request):
+    return request.param
+
+
+@pytest.fixture(params=_DTYPES)
+def dtype(request):
+    return request.param
+
+
+@pytest.fixture(params=list(_EXPECTED_DATE_TYPES.keys()))
+def calendar(request):
+    return request.param
+
+
+@pytest.mark.parametrize("unit", _MICROSECOND_UNITS)
+def test_num2date_microsecond_units(calendar, unit, shape, dtype):
+    date_type = _EXPECTED_DATE_TYPES[calendar]
+    expected = np.array([date_type(2000, 1, 1, 0, 0, 0, 1),
+                         date_type(2000, 1, 1, 0, 0, 0, 2),
+                         date_type(2000, 1, 1, 0, 0, 0, 3),
+                         date_type(2000, 1, 1, 0, 0, 0, 4)]).reshape(shape)
+    numeric_times = np.array([1, 2, 3, 4]).reshape(shape).astype(dtype)
+    units = "{} since 2000-01-01".format(unit)
+    result = num2date(numeric_times, units=units, calendar=calendar)
+    np.testing.assert_equal(result, expected)
+
+
+@pytest.mark.parametrize("unit", _MILLISECOND_UNITS)
+def test_num2date_millisecond_units(calendar, unit, shape, dtype):
+    date_type = _EXPECTED_DATE_TYPES[calendar]
+    expected = np.array([date_type(2000, 1, 1, 0, 0, 0, 1000),
+                         date_type(2000, 1, 1, 0, 0, 0, 2000),
+                         date_type(2000, 1, 1, 0, 0, 0, 3000),
+                         date_type(2000, 1, 1, 0, 0, 0, 4000)]).reshape(shape)
+    numeric_times = np.array([1, 2, 3, 4]).reshape(shape).astype(dtype)
+    units = "{} since 2000-01-01".format(unit)
+    result = num2date(numeric_times, units=units, calendar=calendar)
+    np.testing.assert_equal(result, expected)
+
+
+@pytest.mark.parametrize("unit", _SECOND_UNITS)
+def test_num2date_second_units(calendar, unit, shape, dtype):
+    date_type = _EXPECTED_DATE_TYPES[calendar]
+    expected = np.array([date_type(2000, 1, 1, 0, 0, 1, 0),
+                         date_type(2000, 1, 1, 0, 0, 2, 0),
+                         date_type(2000, 1, 1, 0, 0, 3, 0),
+                         date_type(2000, 1, 1, 0, 0, 4, 0)]).reshape(shape)
+    numeric_times = np.array([1, 2, 3, 4]).reshape(shape).astype(dtype)
+    units = "{} since 2000-01-01".format(unit)
+    result = num2date(numeric_times, units=units, calendar=calendar)
+    np.testing.assert_equal(result, expected)
+
+
+@pytest.mark.parametrize("unit", _MINUTE_UNITS)
+def test_num2date_minute_units(calendar, unit, shape, dtype):
+    date_type = _EXPECTED_DATE_TYPES[calendar]
+    expected = np.array([date_type(2000, 1, 1, 0, 1, 0, 0),
+                         date_type(2000, 1, 1, 0, 2, 0, 0),
+                         date_type(2000, 1, 1, 0, 3, 0, 0),
+                         date_type(2000, 1, 1, 0, 4, 0, 0)]).reshape(shape)
+    numeric_times = np.array([1, 2, 3, 4]).reshape(shape).astype(dtype)
+    units = "{} since 2000-01-01".format(unit)
+    result = num2date(numeric_times, units=units, calendar=calendar)
+    np.testing.assert_equal(result, expected)
+
+
+@pytest.mark.parametrize("unit", _HOUR_UNITS)
+def test_num2date_hour_units(calendar, unit, shape, dtype):
+    date_type = _EXPECTED_DATE_TYPES[calendar]
+    expected = np.array([date_type(2000, 1, 1, 1, 0, 0, 0),
+                         date_type(2000, 1, 1, 2, 0, 0, 0),
+                         date_type(2000, 1, 1, 3, 0, 0, 0),
+                         date_type(2000, 1, 1, 4, 0, 0, 0)]).reshape(shape)
+    numeric_times = np.array([1, 2, 3, 4]).reshape(shape).astype(dtype)
+    units = "{} since 2000-01-01".format(unit)
+    result = num2date(numeric_times, units=units, calendar=calendar)
+    np.testing.assert_equal(result, expected)
+
+
+@pytest.mark.parametrize("unit", _DAY_UNITS)
+def test_num2date_day_units(calendar, unit, shape, dtype):
+    date_type = _EXPECTED_DATE_TYPES[calendar]
+    expected = np.array([date_type(2000, 1, 2, 0, 0, 0, 0),
+                         date_type(2000, 1, 3, 0, 0, 0, 0),
+                         date_type(2000, 1, 4, 0, 0, 0, 0),
+                         date_type(2000, 1, 5, 0, 0, 0, 0)]).reshape(shape)
+    numeric_times = np.array([1, 2, 3, 4]).reshape(shape).astype(dtype)
+    units = "{} since 2000-01-01".format(unit)
+    result = num2date(numeric_times, units=units, calendar=calendar)
+    np.testing.assert_equal(result, expected)
+
+
+@pytest.mark.parametrize("unit", _MONTH_UNITS)
+def test_num2date_month_units(calendar, unit, shape, dtype):
+    date_type = _EXPECTED_DATE_TYPES[calendar]
+    expected = np.array([date_type(2000, 2, 1, 0, 0, 0, 0),
+                         date_type(2000, 3, 1, 0, 0, 0, 0),
+                         date_type(2000, 4, 1, 0, 0, 0, 0),
+                         date_type(2000, 5, 1, 0, 0, 0, 0)]).reshape(shape)
+    numeric_times = np.array([1, 2, 3, 4]).reshape(shape).astype(dtype)
+    units = "{} since 2000-01-01".format(unit)
+
+    if calendar != "360_day":
+        with pytest.raises(ValueError):
+            num2date(numeric_times, units=units, calendar=calendar)
+    else:
+        result = num2date(numeric_times, units=units, calendar=calendar)
+        np.testing.assert_equal(result, expected)
+
+
+def test_num2date_only_use_python_datetimes(calendar, shape, dtype):
+    date_type = real_datetime
+    expected = np.array([date_type(2000, 1, 2, 0, 0, 0, 0),
+                         date_type(2000, 1, 3, 0, 0, 0, 0),
+                         date_type(2000, 1, 4, 0, 0, 0, 0),
+                         date_type(2000, 1, 5, 0, 0, 0, 0)]).reshape(shape)
+    numeric_times = np.array([1, 2, 3, 4]).reshape(shape).astype(dtype)
+    units = "days since 2000-01-01"
+    if calendar not in _STANDARD_CALENDARS:
+        with pytest.raises(ValueError):
+            num2date(numeric_times, units=units, calendar=calendar,
+                           only_use_python_datetimes=True,
+                           only_use_cftime_datetimes=False)
+    else:
+        result = num2date(numeric_times, units=units, calendar=calendar,
+                                only_use_python_datetimes=True,
+                                only_use_cftime_datetimes=False)
+        np.testing.assert_equal(result, expected)
+
+
+def test_num2date_use_pydatetime_if_possible(calendar, shape, dtype):
+    if calendar not in _STANDARD_CALENDARS:
+        date_type = _EXPECTED_DATE_TYPES[calendar]
+    else:
+        date_type = real_datetime
+
+    expected = np.array([date_type(2000, 1, 2, 0, 0, 0, 0),
+                         date_type(2000, 1, 3, 0, 0, 0, 0),
+                         date_type(2000, 1, 4, 0, 0, 0, 0),
+                         date_type(2000, 1, 5, 0, 0, 0, 0)]).reshape(shape)
+    numeric_times = np.array([1, 2, 3, 4]).reshape(shape).astype(dtype)
+    units = "days since 2000-01-01"
+    result = num2date(numeric_times, units=units, calendar=calendar,
+                            only_use_python_datetimes=False,
+                            only_use_cftime_datetimes=False)
+    np.testing.assert_equal(result, expected)
+
+
+@pytest.mark.parametrize(
+    ["standard_calendar", "breakpoint"],
+    [("proleptic_gregorian", "{}-12-31T23:59:59.999999".format(MINYEAR - 1)),
+     ("gregorian", "1582-10-15"),
+     ("standard", "1582-10-15")]
+)
+def test_num2date_only_use_python_datetimes_invalid_basedate(
+    standard_calendar,
+    breakpoint
+):
+    units = "days since {}".format(breakpoint)
+    numeric_times = np.array([1, 2, 3, 4])
+    with pytest.raises(ValueError):
+        num2date(
+            numeric_times,
+            units=units,
+            calendar=standard_calendar,
+            only_use_python_datetimes=True,
+            only_use_cftime_datetimes=False
+        )
+
+
+@pytest.mark.parametrize("real_world_calendar", _REAL_WORLD_CALENDARS)
+def test_num2date_invalid_zero_reference_year(real_world_calendar):
+    units = "days since 0000-01-01"
+    numeric_times = np.array([1, 2, 3, 4])
+    with pytest.raises(ValueError, match="zero not allowed as a reference"):
+        num2date(
+            numeric_times,
+            units=units,
+            calendar=real_world_calendar
+        )
+
+
+@pytest.mark.parametrize("artificial_calendar", _ARTIFICIAL_CALENDARS)
+def test_num2date_valid_zero_reference_year(artificial_calendar):
+    units = "days since 0000-01-01"
+    numeric_times = np.array([1, 2, 3, 4])
+    num2date(numeric_times, units=units, calendar=artificial_calendar)
+
+
+def test_num2date_masked_array(calendar):
+    date_type = _EXPECTED_DATE_TYPES[calendar]
+    expected = np.array([date_type(2000, 1, 1, 1, 0, 0, 0),
+                         date_type(2000, 1, 1, 2, 0, 0, 0),
+                         date_type(2000, 1, 1, 3, 0, 0, 0),
+                         date_type(2000, 1, 1, 4, 0, 0, 0)])
+    mask = [False, False, True, False]
+    expected = np.ma.masked_array(expected, mask=mask)
+    numeric_times = np.ma.masked_array([1, 2, 3, 4], mask=mask)
+    units = "hours since 2000-01-01"
+    result = num2date(numeric_times, units=units, calendar=calendar)
+    np.testing.assert_equal(result, expected)
+
+
+def test_num2date_out_of_range():
+    numeric_times = 12 * np.array([200000, 400000, 600000])
+    units = "months since 2000-01-01"
+    with pytest.raises(OverflowError, match="time values outside range of 64 bit signed integers"):
+        num2date(numeric_times, units=units, calendar="360_day")
+
+
+def test_num2date_list_input(calendar):
+    date_type = _EXPECTED_DATE_TYPES[calendar]
+    expected = np.array([date_type(2000, 1, 1, 1, 0, 0, 0),
+                         date_type(2000, 1, 1, 2, 0, 0, 0),
+                         date_type(2000, 1, 1, 3, 0, 0, 0),
+                         date_type(2000, 1, 1, 4, 0, 0, 0)])
+    numeric_times = [1, 2, 3, 4]
+    units = "hours since 2000-01-01"
+    result = num2date(numeric_times, units=units, calendar=calendar)
+    np.testing.assert_equal(result, expected)
+
+
+def test_num2date_integer_upcast_required():
+    numeric_times = np.array([30, 60, 90, 120], dtype=np.int32)
+    units = "minutes since 2000-01-01"
+    expected = np.array([
+        Datetime360Day(2000, 1, 1, 0, 30, 0),
+        Datetime360Day(2000, 1, 1, 1, 0, 0),
+        Datetime360Day(2000, 1, 1, 1, 30, 0),
+        Datetime360Day(2000, 1, 1, 2, 0, 0)
+    ])
+    result = num2date(numeric_times, units=units, calendar="360_day")
+    np.testing.assert_equal(result, expected)
+
+
+@pytest.mark.parametrize(
+    "encoding_units", 
+    ["microseconds", "milliseconds", "seconds", "minutes", "hours", "days"]
+)
+@pytest.mark.parametrize(
+    "freq",
+    [
+        timedelta(microseconds=1),
+        timedelta(microseconds=1000), 
+        timedelta(seconds=1), 
+        timedelta(minutes=1), 
+        timedelta(hours=1), 
+        timedelta(days=1)
+    ], 
+    ids=lambda x: f"{x!r}"
+)
+def test_date2num_num2date_roundtrip(encoding_units, freq, calendar):
+    date_type = _EXPECTED_DATE_TYPES[calendar]
+    lengthy_timedelta = timedelta(days=291000 * 360)
+    times = np.array(
+        [
+            date_type(1, 1, 1), 
+            date_type(1, 1, 1) + lengthy_timedelta, 
+            date_type(1, 1, 1) + lengthy_timedelta + freq
+        ]
+    )
+    units = f"{encoding_units} since 0001-01-01"
+    encoded = date2num(times, units=units, calendar=calendar)
+    decoded = num2date(encoded, units=units, calendar=calendar)
+    encoding_units_as_timedelta = timedelta(microseconds=UNIT_CONVERSION_FACTORS[encoding_units])
+
+    if freq >= encoding_units_as_timedelta:
+        assert encoded.dtype == np.int64
+        np.testing.assert_equal(decoded, times)
+    else:
+        assert encoded.dtype == np.float64
+        tolerance = timedelta(microseconds=2000)
+        meets_tolerance = np.abs(decoded - times) <= tolerance
+        assert np.all(meets_tolerance)
+
+
 if __name__ == '__main__':
     unittest.main()
